/**
 * Base class for all services
 * for service discovery
 * @memberof module:core
 */
class AbstractService {
  /**
   * @constructor
   * @param {String} servicePath - path to service implementation. It could be
   *        path to the file or node_modules package. For bundled services use
   *        `microset/` prefix (e.g. `microset/EchoService`)
   * @param {Array} labels - list of labels that can be used for querying.
   *        You can provide a String or Array of Strings here
   * @param {ServiceRegistry} serviceRegistry - registry where the service is
   *        stored
   */
  constructor(servicePath, labels, serviceRegistry) {
    this._serviceRegistry = serviceRegistry;
    labels = labels || [];
    if (typeof(labels) == 'object') {
      labels = Object.values(labels);
    } else {
      labels = [String(labels)];
    }
    this._servicePath = servicePath;
    this._labels = labels;
  }

  /**
   * Unique ID of service instance
   * @type {Array}
   */
  get id() {
    throw new Error('Not implemented');
  }

  /**
   * Unique ID of service instance
   * @type {Array}
   */
  get name() {
    throw new Error('Not implemented');
  }

  /**
   * Type of the service (e.g. `hosted`, `native`, `remote`)
   * @type {String}
   */
  get type() {
    throw new Error('Not implemented');
  }

  /**
   * Determines whether the service was initialised already or not.
   * @type {Boolean}
   */
  get isReady() {
    throw new Error('Not implemented');
  }

  /**
   * Determines whether the service is running and online (e.g. the process is
   * running)
   * @type {Boolean}
   */
  get isOnline() {
    throw new Error('Not implemented');
  }

  /**
   * Adds event listener waiting for service to be ready
   * @param {Function} callback - function to be called when the service is
   *        ready. Service wrapper reference is passed to the callback
   *        as an argument.
   */
  onReady(callback) {
    throw new Error('Not implemented');
  }

  /**
   * removes event listener waiting for service to be ready
   * @param {Function} callback - function to be called when the service is
   *        ready.
   */
  offReady(callback) {
    throw new Error('Not implemented');
  }

  /**
   * Check if a method is exposed by wrapped service
   * @param {string} methodName - name of method to be checked.
   * @return {Boolean} true if the method is exposed, otherwise false
   */
  isExposed(methodName) {
    throw new Error('Not implemented');
    return false;
  }

  /**
   * @return {Array} list of exposed methods
   */
  getExposedMethods() {
    throw new Error('Not implemented');
    return [];
  }

  /**
   * Executes an exposed method of the service
   * @param {string} methodName - name of method to be called.
   * @param {Array} params - list of parameters to the method.
   * @return {Object} response from the service
   */
  async call(methodName, params) {
    throw new Error('Not implemented');
  }

  /**
   * List of labels assigned to the
   * NOTICE: singular form is used for easier querying!
   * @private
   * @type {Array}
   */
  get label() {
    return this._labels;
  }

  /**
   * @return {String} path to service implementation. It could be
   *         path to the file or node_modules package. For bundled services use
   *         `microset/` prefix (e.g. `microset/EchoService`)
   */
  getServicePath() {
    return this._servicePath;
  }

  /**
   * @return {Array} list of labels that can be used for querying.
   *         You can provide a String or Array of Strings here
   */
  getLabels() {
    return this._labels;
  }

  /**
   * @return {String} unique service id within namespace of related service
   *         registry
   */
  generateServiceId() {
    return this._serviceRegistry.id +
      '.' +
      Math.floor(Math.random()*0xffffffff).toString(16);
  }

  /**
   * @return {Object} JSON describing properties of the object
   */
  describe() {
    return {
      id: this.id,
      name: this.name,
      type: this.type,
      ready: this.isReady,
      online: this.isOnline,
      methods: this.getExposedMethods(),
      labels: this.getLabels(),
    };
  }

  /**
   * check if the service match query criteria
   * @param {String} query - query to select the service. For example:
   *        MyServiceName, MyServiceName?label=dev, ?label=dev
   * @return {Boolean} - true if service match the criteria, otherwise false
   */
  match(query) {
    query = query.split('?');
    const nameQuery = query[0];
    let paramQuery = query.length > 1 ? query[1] : '';

    if (!nameQuery && !paramQuery) {
      return false;
    }

    if (nameQuery && this.name != nameQuery) {
      return false;
    }

    paramQuery = paramQuery == '' ? [] : paramQuery
        .split('&')
        .map((param) => {
          param = param.split('=');
          const key = param[0];
          let value = param.length > 1 ? param[1] : true;
          value = value === 'true' ? true : value;
          value = value === 'false' ? false : value;
          return {key: key, value: value};
        });

    // TODO: add whitelisting
    for (const param of paramQuery) {
      if (Array.isArray(this[param.key])) {
        if (this[param.key].indexOf(param.value) == -1) {
          return false;
        }
      } else {
        if (this[param.key] !== param.value) {
          return false;
        }
      }
    }
    return true;
  }
}

module.exports = AbstractService;
