const ProcessRpcBridge = require('./process/ProcessRpcBridge.js');
const ServiceProxy = require('./process/ServiceProxy.js');
const AbstractService = require('./AbstractService.js');
const path = require('path');
const childProcess = require('child_process');

/**
 * All custom services are hosted. It means that they run in separate processes
 * and are managed by the registry.
 * The class is a wrapper for service logic. It initialise the service and
 * expose its selected methods. It also proxy all micro service calls
 * to provide proper logging and error handling
 * @memberof module:core
 */
class HostedService extends AbstractService {
  /**
   * @constructor
   * @param {ServiceRegistry} serviceRegistry - registry where the service is
   *        stored
   * @param {Logger} logger - logger instance
   * @param {String} servicePath - path to service implementation. It could be
   *        path to the file or node_modules package. For bundled services use
   *        `microset/` prefix (e.g. `microset/EchoService`)
   * @param {Array} labels - list of labels that can be used for querying.
   *        You can provide a String or Array of Strings here
   */
  constructor(serviceRegistry, logger, servicePath, labels) {
    super(servicePath, labels, serviceRegistry);
    this._logger = logger;
    this._isReady = false;
    this._onReadyCallbacks = [];
    this._exposedMethods = [];
    this._serviceId = null;
    this._serviceName = null;
    this._bridge = null;
    this._isOnline = false;
    this._resolveCreate = null;
    this._proxy = new ServiceProxy(this, this._serviceRegistry);
    this._watchdog = null;
    this._watchdogInterval = 10000;
    this._restartJob = null;
    this._keepRunning = true;
    this._restartTimeout = 3000;
    this._config = {};
  }

  /**
   * Type of the service (e.g. `hosted`, `native`, `remote`)
   * @type {String}
   */
  get type() {
    return 'hosted';
  }

  /**
   * configuration of the service
   * @type {Object}
   */
  get config() {
    return JSON.parse(JSON.stringify(this._config));
  }

  /**
   * Verify if service module exists at given path
   * @private
   * @param {String} servicePath - path to service implementation. It could be
   *        path to the file or node_modules package. For bundled services use
   *        `microset/` prefix (e.g. `microset/EchoService`)
   * @return {Boolean} true if service exists
   */
  _moduleExists(servicePath) {
    try {
      require.resolve(servicePath);
    } catch (err) {
      return false;
    }
    return true;
  }

  /**
   * @private
   * @param {String} servicePath - path to service implementation. It could be
   *        path to the file or node_modules package. For bundled services use
   *        `microset/` prefix (e.g. `microset/EchoService`)
   * @return {String} absolute path to service module
   */
  _resolveServicePath(servicePath) {
    const bundledServices = {
      'microset/EchoService': 'services/echo/EchoService.js',
      'microset/HttpServerService': 'services/httpServer/HttpServerService.js',
      'microset/JsonRpcGatewayService': 'services/jsonRpcGateway/JsonRpcGatewayService.js',
      'microset/JsonRpcClientService': 'services/jsonRpcClient/JsonRpcClientService.js',
      'microset/JsonStorageService': 'services/jsonStorage/JsonStorageService.js',
      'microset/StaticContentService': 'services/staticContent/StaticContentService.js',
    };
    if (bundledServices[servicePath] !== undefined) {
      return path.resolve(__dirname, bundledServices[servicePath]);
    }
    if (!path.isAbsolute(servicePath)) {
      servicePath = path.resolve(process.cwd(), servicePath);
    }
    if (this._moduleExists(servicePath)) {
      return servicePath;
    }
    throw new Error(`Unable to find service file at '${servicePath}'`);
  }

  /**
   * @private
   * notify that service was created
   */
  async onProcessCreated() {
    this._logger.info(`SERVICE ID: ${this._serviceId}`);
    this._resolveCreate();
    this._resolveCreate = null;
  }

  /**
   * Launch process that will host service implementation
   * @async
   * @param {String} servicePath - path to service calss. It could be also any
   *        node_module. For bundled services use microset prefix
   *        (e.g. microset/EchoService)
   * @return {Promise}
   */
  create() {
    if (this._resolveCreate) {
      return Promise.reject(new Error('Service creation requested already!'));
    }
    return new Promise((resolve, reject) => {
      this._resolveCreate = resolve;
      let servicePath = this.getServicePath();
      servicePath = this._resolveServicePath(servicePath);

      this._serviceName = servicePath.replace(/^(.*[\\\/])?([^\.\\\/]*)(\.js)?$/, '$2');
      this._logger = this._logger.getLogger(this._serviceName);
      this._serviceId = this.generateServiceId();

      const serviceProcess = childProcess.spawn(
          'node',
          [
            path.resolve(__dirname, 'process', 'serviceLauncher.js'),
            servicePath,
          ]
      );
      this._isOnline = true;
      this._bridge = new ProcessRpcBridge();
      this._bridge.attachToProcess(serviceProcess, 'stdout', 'stdin');
      this._bridge.attachToTarget(this._proxy);

      this._bridge.onClose((code) => {
        code = code || 'unknown';
        this._logger.fatal(`Process of '${this._serviceName}' service closed with code ${code}`);
        this._handleProcessClose(true);
      });

      this._bridge.onError((msg) => {
        this._logger.error(msg);
      });

      if (this._watchdog) {
        clearInterval(this._watchdog);
        this._watchdog = null;
      }
    });
  }

  /**
   * cleanup after closing service process
   * @param {Boolean} unexpected - whether closing of the service was unexpected
   * @private
   */
  _handleProcessClose(unexpected) {
    if (!this._restartJob && this._keepRunning && unexpected) {
      const logger = this._logger;
      logger.warn('Service process is down. Restart in ' + this._restartTimeout + 'ms');
      this._restartJob = setTimeout(() => {
        logger.info('Keeping the service running after down time');
        this._serviceRegistry.reload(this.id);
      }, this._restartTimeout);
    }
    if (this._bridge) {
      this._logger = this._logger.parent;
      this._bridge.close();
      this._bridge = null;
    }
    if (this._watchdog) {
      clearInterval(this._watchdog);
      this._watchdog = null;
    }
    this._isOnline = false;
    this._isReady = false;
    this._resolveCreate = null;
    this._bridge = null;
  }

  /**
   * Close the service destroying related process and going to offline state
   */
  close() {
    if (this._bridge) {
      this._bridge.onClose(() => {});
      this._bridge.close();
      this._logger.debug(`Process of '${this._serviceName}' (ID: '${this._serviceId}') closed`);
    }
    this._handleProcessClose();
  }

  /**
   * Unique ID of service instance
   * @type {String}
   */
  get id() {
    return this._serviceId;
  }

  /**
   * Adds event listener waiting for service to be ready
   * @param {Function} callback - function to be called when the service is
   *        ready. Service reference is passed to the callback
   *        as an argument.
   */
  onReady(callback) {
    this._onReadyCallbacks.push(callback);
  }

  /**
   * removes event listener waiting for service to be ready
   * @param {Function} callback - function to be called when the service is
   *        ready.
   */
  offReady(callback) {
    this._onReadyCallbacks = this._onReadyCallbacks.filter((c) => {
      return c != callback;
    });
  }

  /**
   * Initialise the service.
   * @param {Object} config - configuration passed to the service
   */
  async bootstrap(config) {
    this._config = JSON.parse(JSON.stringify(config));
    const serviceConfig = config.service || {};
    const timeout = serviceConfig.timeout || 5000;
    let keepRunning = serviceConfig.keepRunning;
    keepRunning = (keepRunning === undefined) ? true : keepRunning;
    this._keepRunning = keepRunning;
    this._logger.info(`Initialising ${this.name} service...`);
    try {
      this._bridge.timeout = timeout;
      await this._bridge.call('setConfig', [config]);
      await this._bridge.call('init');
      this._exposedMethods = await this._bridge.call('getExposedMethods');
    } catch (err) {
      this._logger.fatal(`Unable to initialise ${this._serviceName} service: ${err}`);
      if (err.stack) {
        err.stack.split('\n').forEach((row) => this._logger.debug(row));
      }
      return;
    }
    this._logger.info(`Service ${this.name} initialised.`);
    this._exposedMethods.forEach((methodName) => {
      this._logger.info(`Exposing method: '${methodName}(...)'`);
    });
    this._isReady = true;
    this._logger.info(`Service ready`);
    this._onReadyCallbacks.forEach((c) => c(this));

    const watchTime = Math.round(this._watchdogInterval * 0.7 +
                    Math.random() * this._watchdogInterval * 0.6);
    this._watchdog = setInterval(this._watchService.bind(this), watchTime);
  }

  /**
   * Executes an exposed method of the service
   * @param {string} methodName - name of method to be called.
   * @param {Array} params - list of parameters to the method.
   * @return {Object} response from the service
   */
  async call(methodName, params) {
    if (!Array.isArray(params)) {
      throw new Error(`Cannot call ${methodName}'. Parameters must be an array`);
    }
    if (!this.isExposed(methodName)) {
      throw new Error(`Method '${this._serviceName}.${methodName}' not found`);
    }
    if (!this._isOnline) {
      throw new Error(`Service '${this._serviceName}' is offline`);
    }
    this._logger.debug(`Calling ${this._serviceName}.${methodName}()...`);
    let result;
    try {
      result = await this._bridge.call(methodName, params);
    } catch (err) {
      this._logger.error(`Error from method ${this._serviceName}.${methodName}(): ${err.message}`);
      if (err.stack) {
        err.stack.split('\n').forEach((row) => this._logger.debug(row));
      };
      throw err;
    }
    this._logger.debug(`Response from ${this._serviceName}.${methodName}() received`);
    return result;
  }

  /**
   * making regular health checks of the service and close it if not responding
   * @private
   */
  async _watchService() {
    try {
      const pong = await this._bridge.call('ping');
      if (pong !== 'pong') throw new Error(`Unexpected ping response: '${pong}'`);
    } catch (err) {
      this._logger.fatal(`Service ${this._serviceName} (ID: ${this._serviceId}) not responding: ` + String(err));
      if (this._bridge) {
        this._bridge.onClose(() => {});
        this._logger.debug(`Process of '${this._serviceName}' (ID: '${this._serviceId}') closed`);
      }
      this._handleProcessClose(true);
    }
  }

  /**
   * Check if a method is exposed by the service
   * @param {string} methodName - name of method to be checked.
   * @return {Boolean} true if the method is exposed, otherwise false
   */
  isExposed(methodName) {
    return this._exposedMethods.indexOf(methodName) !== -1;
  }

  /**
   * @return {Array} list of exposed methods
   */
  getExposedMethods() {
    return this._exposedMethods;
  }

  /**
   * Determines whether the service was initialised already or not.
   * @type {Boolean}
   */
  get isReady() {
    return this._isReady;
  }

  /**
   * Determines whether the service is running and online (e.g. the process is
   * running)
   * @type {Boolean}
   */
  get isOnline() {
    return this._isOnline;
  }

  /**
   * Logger assigned to the service
   * @type {Logger}
   */
  get logger() {
    return this._logger;
  }

  /**
   * Name of the service. Generate automaticaly from service class name
   * @type {String}
   */
  get name() {
    return this._serviceName;
  }
};

module.exports = HostedService;
