/**
 * Logs and routes messages from microset
 * @memberof module:core
 */
class Logger {
  /**
   * @constructor
   * @param {String} name - name of the logger. In most cases this is the
   *        module name that needs to be appended to log message. Default
   *        value is Core
   * @param {Logger} parent - optional parent logger. When provided, logging
   *        output mechanism of parent will be used instead. Could be usefull
   *        to have the same output from many, connected loggers
   */
  constructor(name, parent) {
    this._name = name || 'Core';
    this._parent = parent || null;
    this._logFunction = console.log;
    this._color = true;

    this._levelList = ['debug', 'info', 'warn', 'error', 'fatal'];
    this._level = 'info';

    this._colorPrefix = {
      fatal: '\x1b[7m\x1b[1m\x1b[31m',
      error: '\x1b[1m\x1b[31m',
      warn: '\x1b[1m\x1b[33m',
      info: '\x1b[37m',
      debug: '\x1b[2m',
    };
  }

  /**
   * Determine whether logger output should be colored or not
   * @type {Booelan}
   */
  get color() {
    return this._color;
  }

  // eslint-disable-next-line require-jsdoc
  set color(value) {
    this._color = value;
  }

  /**
   * Threshold for log messages. Only that and higher levels will be logged.
   * available values are: 'debug', 'info', 'warn', 'error', 'fatal'
   * @type {String}
   */
  get level() {
    return this._level;
  }

  /**
   * Name of the logger. It will be appended to log message and can be used
   * to differentiate logs from various modules
   * @type {String}
   */
  get name() {
    return this._name;
  }

  // eslint-disable-next-line require-jsdoc
  set level(value) {
    value = value.toLowerCase();
    if (this._levelList.indexOf(value) != -1) {
      this._level = value;
    }
  }

  /**
   * Creates child logger with provided name
   * @param {String} name - name for child logger. When not provided,
   *        name of parent logger will be used
   * @return {Logger} child logger
   */
  getLogger(name) {
    return new Logger(name || this._name, this);
  }

  /**
   * Outputs log messages
   * @param {String} name - name of the logger
   * @param {String} level - level of the message
   * @param {String} message - message to be output
   * @return {Null}
   * @private
   */
  _log(name, level, message) {
    if (this._parent) {
      return this._parent._log(name, level, message);
    }
    message = String(message);

    const levelThreshold = this._levelList.indexOf(this._level);
    const levelIndex = this._levelList.indexOf(level);
    if (levelIndex < levelThreshold) {
      return;
    }

    const padRight = (txt, len) => {
      while (txt.length < len) {
        txt = txt + ' ';
      }
      return txt;
    };
    const now = (new Date()).toISOString().replace(/[TZ]/g, ' ');
    let msg = '';
    if (this._color) {
      msg += this._colorPrefix[level];
    }
    msg += `${now}|`;
    msg += ` ${padRight(name, 30)}|`;
    msg += ` ${padRight(level.toUpperCase(), 6)}|`;
    msg += ` ${message}`;
    if (this._color) {
      msg += '\x1b[0m';
    }
    this._logFunction(msg);
  }

  /**
   * Logs debug level message
   * @param {String} msg - message to be logged
   */
  debug(msg) {
    this._log(this._name, 'debug', msg);
  }

  /**
   * Logs info level message
   * @param {String} msg - message to be logged
   */
  info(msg) {
    this._log(this._name, 'info', msg);
  }

  /**
   * Logs warning level message
   * @param {String} msg - message to be logged
   */
  warn(msg) {
    this._log(this._name, 'warn', msg);
  }

  /**
   * Logs error level message
   * @param {String} msg - message to be logged
   */
  error(msg) {
    this._log(this._name, 'error', msg);
  }

  /**
   * Logs fatal level message
   * @param {String} msg - message to be logged
   */
  fatal(msg) {
    this._log(this._name, 'fatal', msg);
  }

  /**
   * Parent logger. All log messages will be passed there. It could be usefull
   * to agregate output from many, connected loggers that has differnt settings
   * @type {Logger}
   */
  get parent() {
    return this._parent;
  }
};

module.exports = Logger;
