const AbstractService = require('./AbstractService.js');

/**
 * Native service running within the same process as the registry. Has
 * access to all native features
 * @memberof module:core
 */
class NativeService extends AbstractService {
  /**
   * @constructor
   * @param {ServiceRegistry} serviceRegistry - registry where the service is
   *        stored
   */
  constructor(serviceRegistry) {
    super('', [], serviceRegistry);
    this._exposedMethods = [];
    this._serviceName = this.constructor.name;
    this._serviceId = this.generateServiceId();
    this._logger = serviceRegistry.logger.getLogger(this._serviceName);
    this._logger.info(`SERVICE ID: ${this._serviceId}`);
    this.init();
    this._logger.info(`Service ${this._serviceName} initialised.`);
    this._logger.info(`Service ready`);
  }

  /**
   * Mark a method of the object as to be exposed externally and accessible for
   * other services
   * @param {String} methodName - name of the method to be exposed.
   */
  expose(methodName) {
    this._exposedMethods.push(methodName);
  }

  /**
   * Initializes the service. All mehods that should be available
   * externally should be exposed here by calling this.expose(methodName).
   */
  init() {

  }

  /**
   * Unique ID of service instance
   * @type {Array}
   */
  get name() {
    return this._serviceName;
  }

  /**
   * Type of the service (e.g. `hosted`, `native`, `remote`)
   * @type {String}
   */
  get type() {
    return 'native';
  }

  /**
   * Unique ID of service instance
   * @type {Array}
   */
  get id() {
    return this._serviceId;
  }

  /**
   * Adds event listener waiting for service to be ready
   * @param {Function} callback - function to be called when the service is
   *        ready. Service wrapper reference is passed to the callback
   *        as an argument.
   */
  onReady(callback) {

  }

  /**
   * removes event listener waiting for service to be ready
   * @param {Function} callback - function to be called when the service is
   *        ready.
   */
  offReady(callback) {

  }

  /**
  * Determines whether the service was initialised already or not.
  * @type {Boolean}
  */
  get isReady() {
    return true;
  }

  /**
  * Determines whether the service is running and online (e.g. the process is
  * running)
  * @type {Boolean}
  */
  get isOnline() {
    return true;
  }

  /**
   * Executes an exposed method of the service
   * @param {string} methodName - name of method to be called.
   * @param {Array} params - list of parameters to the method.
   * @return {Object} response from the service
   */
  async call(methodName, params) {
    if (!Array.isArray(params)) {
      throw new Error(`Cannot call ${methodName}'. Parameters must be an array`);
    }
    if (!this.isExposed(methodName)) {
      throw new Error(`Method '${this._serviceName}.${methodName}' not found`);
    }
    this._logger.debug(`Calling ${this._serviceName}.${methodName}()...`);
    let result;
    try {
      result = await this[methodName](...params);
    } catch (err) {
      this._logger.error(`Error from method ${this._serviceName}.${methodName}(): ${err.message}`);
      if (err.stack) {
        err.stack.split('\n').forEach((row) => this._logger.debug(row));
      };
      throw err;
    }
    this._logger.debug(`Response from ${this._serviceName}.${methodName}() received`);
    return result;
  }

  /**
  * Check if a method is exposed by wrapped service
  * @param {string} methodName - name of method to be checked.
  * @return {Boolean} true if the method is exposed, otherwise false
  */
  isExposed(methodName) {
    return this._exposedMethods.indexOf(methodName) !== -1;
  }

  /**
   * @return {Array} list of exposed methods
   */
  getExposedMethods() {
    return this._exposedMethods;
  }
}

module.exports = NativeService;
