/* istanbul ignore next */
/**
 * Parent class for hosted micro services. Each service implementation must
 * extend it. It provides the service with: initialization mechanism,
 * service dependency resolution, configuration, exposing of methods,
 * logger.
 * @memberof module:core
 */
class Service {
  /**
   * @constructor
   * @param {ProcessRpcBridge} bridge - Process RPC Bridge used to communicate
   *        with microset core
   */
  constructor(bridge) {
    this._bridge = bridge;
    this._config = {};
    this._exposedMethods = [];
    this._id = undefined;
    this._name = undefined;
  }

  /**
   * Unique ID of service instance
   * @type {String}
   */
  get id() {
    return this._id;
  }

  // eslint-disable-next-line require-jsdoc
  set id(value) {
    this._id = value;
  }

  /**
   * Name of the service
   * @type {String}
   */
  get name() {
    return this._name;
  }

  // eslint-disable-next-line require-jsdoc
  set name(value) {
    this._name = value;
  }

  /**
   * Initializes the service. All mehods that should be available
   * externally should be exposed here by calling this.expose(methodName).
   * The configuration of the service is available at this stage at
   * this.config.
   */
  async init() {

  }

  /**
   * Gets service that match the criteria, and call its method.
   * If there are multiple services matching the criteria, a random one
   * will be returned.
   * @param {String} serviceName - query to select the service. For example:
   *        MyServiceName, MyServiceName?label=dev, ?label=dev
   * @param {String} methodName - name of method to be called.
   * @param {Array} params - list of parameters to the method.
   * @return {Object} result from the service
   */
  async callService(serviceName, methodName, params) {
    return await this._bridge.call('callService', [serviceName, methodName, params]);
  }

  /**
   * Mark a method of the object as to be exposed externally and accessible for
   * other services
   * @param {String} methodName - name of the method to be exposed.
   */
  expose(methodName) {
    if (this._exposedMethods.indexOf(methodName) != -1) {
      throw new Error(`Method ${methodName} already exposed`);
    }
    this._exposedMethods.push(methodName);
  }

  /**
   * @return {Array} list of method names to be exposed by the service
   */
  getExposedMethods() {
    return this._exposedMethods;
  }

  /**
   * configuration of the service
   * @type {Object}
   */
  get config() {
    return this._config;
  }

  /**
   * @return {Object} configuration of the service
   */
  getConfig() {
    return this._config;
  }

  /**
   * @param {Object} value - configuration of the service
   */
  setConfig(value) {
    this._config = value;
  }

  /**
   * Log message with level fatal
   * @param {String} msg - message to be logged
   */
  logFatal(msg) {
    this._bridge.call('logFatal', [msg]);
  }

  /**
   * Log message with level error
   * @param {String} msg - message to be logged
   */
  logError(msg) {
    this._bridge.call('logError', [msg]);
  }

  /**
   * Log message with level warning
   * @param {String} msg - message to be logged
   */
  logWarn(msg) {
    this._bridge.call('logWarn', [msg]);
  }

  /**
   * Log message with level info
   * @param {String} msg - message to be logged
   */
  logInfo(msg) {
    this._bridge.call('logInfo', [msg]);
  }

  /**
   * Log message with level debug
   * @param {String} msg - message to be logged
   */
  logDebug(msg) {
    this._bridge.call('logDebug', [msg]);
  }

  /**
   * checks if the service is responding
   * @return {String} pong
   */
  async ping() {
    return 'pong';
  }
};

module.exports = Service;
