const HostedService = require('./HostedService.js');
const Logger = require('./Logger.js');
const ServiceRouter = require('./ServiceRouter.js');
const DirectoryService = require('./nativeServices/directory/DirectoryService.js');

/**
 * Keeps references to all services and resolve dependencies
 * @memberof module:core
 */
class ServiceRegistry {
  /**
   * @constructor
   */
  constructor() {
    this._id = Math.floor(Math.random()*0xffffffff).toString(16);
    this._logger = new Logger();
    this._services = [];
    this._router = new ServiceRouter();
    this._router.register(new DirectoryService(this));
  }

  /**
   * Unique ID that will be the namespace of services.
   * Usually assosiated with the container.
   * @type {String}
   */
  get id() {
    return this._id;
  }

  /**
   * @param {String} id - id of service
   * @return {HostedService} service with given id
   */
  _getServiceById(id) {
    const result = this._services.filter((item) => item.id == id);
    if (result.length < 1) {
      return null;
    }
    return result[0];
  }

  /**
   * Logger assigned to the service
   * @type {Logger}
   */
  get logger() {
    return this._logger;
  }

  /**
   * Adds a hosted service to the register
   * @param {String} servicePath - path to service calss. It could be also any
   *        node_module. For bundled services use microset prefix
   *        (e.g. microset/EchoService)
   * @param {Object} config - configuration of service that is
   *        passed to it before initialization
   * @param {Array} labels - list of labels that can be used for quiering.
   *        You can provide a String or Array of Strings here
   */
  async register(servicePath, config, labels) {
    labels = labels || [];
    if (config && typeof(config) != 'object') {
      throw new Error('ServiceRegistry: Config must be an object');
    }
    config = config || {};
    const service = new HostedService(
        this,
        this._logger,
        servicePath,
        labels
    );
    try {
      await service.create();
      this._services.push(service);
      service.bootstrap(config); // no await to speed up bootstrapping
      this._router.register(service);
    } catch (err) {
      this._logger.fatal(`Unable to register service ${servicePath}. ${err}`);
    }
    return service.id;
  }

  /**
   * Reload provided service starting it agiain
   * @private
   * @param {String} serviceId - id of service to be reloaded
   * @param {Object} [config] - Optional configuration object. If not provided
   *        current configuration will be used
   */
  async reload(serviceId, config) {
    const service = this._getServiceById(serviceId);
    if (!service) {
      throw new Error(`Unable to reload service '${serviceId}'. Service not found!`);
    }
    this._logger.info(`Reloading '${service.name}' (ID: ${service.id})`);
    await this.unregister(serviceId);

    config = config || service.config;
    await this.register(
        service.getServicePath(),
        config,
        service.getLabels()
    );
    this._logger.info(`Service '${service.name}' (ID: ${service.id}) reloaded`);
  }

  /**
   * Unregisters provided service and stops it
   * @private
   * @param {String} serviceId - id of service to be reloaded
   */
  async unregister(serviceId) {
    const service = this._getServiceById(serviceId);
    if (!service) {
      throw new Error(`Unable to unregister service '${serviceId}'. Service not found!`);
    }
    this._logger.info(`Unregistering '${service.name}' (ID: ${service.id})`);
    this._services = this._services.filter((item) => {
      return (item !== service);
    });
    this._router.unregister(service);
    service.close();
    this._logger.info(`Service '${service.name}' (ID: ${service.id}) unregistered`);
  }

  /**
   * Calls service from the registry
   * @param {String} serviceName - query to select the service. For example:
   *        MyServiceName, MyServiceName?label=dev, ?label=dev
   * @param {String} methodName - name of method to be call
   * @param {Array} params - list of parameters to be passed to the method
   * @return {Object} response from the service
   */
  async callService(serviceName, methodName, params) {
    const service = await this._router.getService(serviceName);

    if (!service) {
      this._logger.debug(`Service '${serviceName}' not found`);
      throw new Error(`Service '${serviceName}' not found`);
    }
    if (!service.isExposed(methodName)) {
      this._logger.debug(`Method '${methodName}' not found in '${serviceName}'`);
      throw new Error(`Method '${methodName}' not found in '${serviceName}'`);
    }
    const result = await service.call(
        methodName,
        Object.values(params)
    );
    return result;
  }

  /**
   * @return {Object} JSON describing properties of the object
   */
  describe() {
    return {
      id: this._id,
      services: this._router.describe(),
    };
  }
};

module.exports = ServiceRegistry;
