/**
 * route traffic/calls to serices
 */
class ServiceRouter {
  /**
   * @constructor
   */
  constructor() {
    this._endpoints = [];
  }

  /**
   * adds service to the router
   * @param {AbstractService} service - service to be added
   */
  register(service) {
    this._endpoints.push(service);
  }

  /**
   * removes service to the router
   * @param {AbstractService} service - service to be removed
   */
  unregister(service) {
    this._endpoints = this._endpoints.filter((item) => {
      return (item !== service);
    });
  }

  /**
   * Immediately gets all service that match the criteria, no matter
   * whether it is in ready state or not
   * @param {String} query - query to select the service. For example:
   *        MyServiceName, MyServiceName?label=dev, ?label=dev
   * @return {Array} array of HostedService objects
   */
  getAllServicesInAnyState(query) {
    return this._endpoints
        .filter((service) => (
          service.match(query) &&
          service.isOnline
        ));
  }

  /**
   * Gets service that match the criteria. If it is not in ready state
   * the function wait until the service is initialized.
   * If there are multiple services matching the criteria, a random one
   * will be returned.
   * @param {String} query - query to select the service. For example:
   *        MyServiceName, MyServiceName?label=dev, ?label=dev
   * @return {AbstractService}
   */
  async getService(query) {
    return await (new Promise((resolve) => {
      const matchedServices = this.getAllServicesInAnyState(query);
      if (matchedServices.length == 0) {
        return resolve(null);
      }
      let randomIndex;
      const readyServices = matchedServices.filter((s) => s.isReady);
      if (readyServices.length > 0) {
        randomIndex = Math.floor(Math.random()*readyServices.length);
        return resolve(readyServices[randomIndex]);
      }
      const callbackList = [];
      for (const service of matchedServices) {
        const callback = () => {
          while (callbackList.length) {
            service.offReady(callbackList.pop());
          }
          return resolve(service);
        };
        callbackList.push(callback);
        service.onReady(callback);
      }
    }));
  }

  /**
   * @return {Object} JSON describing properties of the object
   */
  describe() {
    return this._endpoints.map((service) => service.describe());
  }
};

module.exports = ServiceRouter;
