
/**
 * Starts JSON-RPC gateway for the container
 * @param {ServiceRegistry} serviceRegistry - registry where the service is
 *        stored
 * @param {Object} config - container configuration
 *
 * config example:
 * ```
 * {
 *   http: {
 *     port: 8080,
 *     host: 'localhost',
 *   },
 *   jsonRpc: {
 *     apiPath: '/api',
 *     webClientPath: '',
 *     clientLibPath: '/js/microset_jsonrpc_client.js',
 *     whitelist: [
 *       {service: "ServiceName", method: "MethodName1"},
 *       {service: "ServiceName", method: "MethodName2"}
 *     ]
 *   },
 *   staticContent: {
 *      webRoot: './public'
 *    }
 * }
 * ```
 */
module.exports = async (serviceRegistry, config) => {
  config = config || {};
  const http = config.http || {};
  const jsonRpc = config.jsonRpc || {};
  const staticContent = config.staticContent || {};
  const clientWebPath = jsonRpc.webClientPath || '';
  let clientLibPath = jsonRpc.clientLibPath;
  if (clientLibPath === undefined) {
    clientLibPath = '/js/microset_jsonrpc_client.js';
  }
  const port = http.port || 8080;
  const host = http.host || 'localhost';
  const apiPath = jsonRpc.apiPath || '/api';
  const whitelist = jsonRpc.whitelist || null;
  const webRoot = staticContent.webRoot || null;

  const httpConfig = {
    port: port,
    host: host,
    routing: [
      {
        prefix: apiPath,
        service: 'JsonRpcGatewayService',
        method: 'request',
        dropPrefix: true,
      },
    ],
  };

  if (clientLibPath) {
    httpConfig.routing.push({
      prefix: clientLibPath,
      service: 'JsonRpcClientService',
      method: 'getClientLib',
      dropPrefix: true,
    });
  }

  if (clientWebPath) {
    httpConfig.routing.push({
      prefix: clientWebPath,
      service: 'JsonRpcClientService',
      method: 'getWebClient',
      dropPrefix: true,
    });
  }

  if (webRoot) {
    httpConfig.routing.push({
      prefix: '/',
      service: 'StaticContentService',
      method: 'readFile',
      dropPrefix: true,
    });
  }

  await serviceRegistry.register('microset/HttpServerService', httpConfig);
  await serviceRegistry.register('microset/JsonRpcGatewayService', {whitelist: whitelist});
  if (clientWebPath || clientLibPath) {
    serviceRegistry.register('microset/JsonRpcClientService', {
      defaultGateway: apiPath,
    });
  }
  if (webRoot) {
    await serviceRegistry.register('microset/StaticContentService', {
      webRoot: webRoot,
    });
  }
};
