const readline = require('readline');
const ProcessRpcBridgeError = require('./ProcessRpcBridgeError.js');
/**
 * Provides remote procedure calls between processes. It requires attachement of
 * process to send/recive all calls and an object where all RPC call are mapped.
 * @memberof module:core
 */
class ProcessRpcBridge {
  /**
   * @constructor
   */
  constructor() {
    this._rpcTarget = null;
    this._onRawMessageCallback = () => {};
    this._onCloseCallback = () => {};
    this._onErrorCallback = () => {};
    this._responseMap = [];
    this._rpcTargetAvailableResolve = [];
    this._timeout = 30000;
  }

  // eslint-disable-next-line require-jsdoc
  set timeout(v) {
    this._timeout = v;
  }

  /**
   * Timeout for calls execution. By default 30000ms
   * @type {Number}
   */
  get timeout() {
    return this._timeout;
  }

  /**
   * @private
   * @return {Proimise} A promise that waits for RCP target is assigned
   */
  _waitForRpcTarget() {
    if (this._rpcTarget) {
      return Promise.resolve();
    }
    return new Promise((resolve, reject) => {
      this._rpcTargetAvailableResolve.push(resolve);
    });
  }

  /**
   * resolve promise related with JSON-RPC request
   * @param {Object} json - body of JSON-RPC request
   * @private
   */
  async _processJsonRpcResponse(json) {
    if (!this._responseMap[json.id]) {
      return;
    }
    if (json.error === undefined) {
      this._responseMap[json.id].resolve(json.result);
      this._responseMap[json.id] = undefined;
    } else {
      const err = ProcessRpcBridgeError.fromJson(json);
      err.extendStack(this._responseMap[json.id].stack);
      this._responseMap[json.id].reject(err);
      this._responseMap[json.id] = undefined;
    }
  }

  /**
   * Calls RPC target obejct and store resolve promise so it can be mapped with
   * request later
   * @param {Object} json - body of JSON-RPC request
   * @private
   */
  async _processJsonRpcRequest(json) {
    await this._waitForRpcTarget();
    let errorResponse;
    if (json.jsonrpc != '2.0') {
      errorResponse = {
        jsonrpc: '2.0',
        error: {
          code: -32600,
          message: `JSON-RPC version ${json.jsonrpc} is not supported`,
        },
      };
    } else if (json.method === undefined) {
      errorResponse = {
        jsonrpc: '2.0',
        error: {
          code: -32600,
          message: 'method field is required',
        },
      };
    } else if (this._rpcTarget[json.method] === undefined) {
      errorResponse = {
        jsonrpc: '2.0',
        error: {
          code: -32601,
          message: `Method ${json.method} not found`,
        },
      };
    }
    if (errorResponse) {
      errorResponse.id = json.id;
      this.sendRaw(errorResponse);
      return;
    }
    json.params = json.params || [];

    let response;
    try {
      const result = await this._rpcTarget[json.method](...json.params);
      response = {
        jsonrpc: '2.0',
        result: result,
      };
      response.id = json.id;
      this.sendRaw(response);
    } catch (err) {
      err = ProcessRpcBridgeError.fromError(err);
      this.sendRaw(err.toJsonRpc(json.id));
    }
  }

  /**
   * Attach an object to the bridge that all RPC requests will be map to.
   * For example running `bridge.call('methodName', [1,2,3])` will call
   * `rpcTarget.methodName(1, 2, 3);`
   * @param {Obejct} rpcTarget - target all RPC reqests
   */
  attachToTarget(rpcTarget) {
    this._rpcTarget = rpcTarget;
    while (this._rpcTargetAvailableResolve.length) {
      const resolve = this._rpcTargetAvailableResolve.pop();
      resolve();
    }
  }

  /**
   * destroys process of the service
   */
  close() {
    if (this._process) {
      this._process.kill();
    }
    for (const i in this._responseMap) {
      if (this._responseMap[i]) {
        this._responseMap[i].reject(new Error(`Service offline`));
      }
    }
    this._responseMap = [];
    this._process = null;
  }

  /**
   * Connect the bridge to a process and defines input and output streams that
   * should be used for communication
   * @param {Process} process - process to communicate with
   * @param {String} [inputName] - `stdin` by default. A stream where all calls
   *        should be sent
   * @param {String} [outputName] - `stdout` by default. A stream from where all
   *        call are going to be received
   */
  attachToProcess(process, inputName, outputName) {
    this._process = process;
    this._inputName = inputName || 'stdin';
    this._outputName = outputName || 'stdout';

    this._process[this._inputName].setEncoding('utf-8');
    this._reader = readline.createInterface({
      input: this._process[this._inputName],
    });

    this._reader.on('line', async (line) => {
      const buff = Buffer.from(line, 'base64');
      const jsonText = buff.toString('ascii');
      // TODO: add error handling
      const json = JSON.parse(jsonText);
      this._onRawMessageCallback(json);

      if (json.jsonrpc !== undefined && json.method !== undefined) {
        await this._processJsonRpcRequest(json);
      } else if (json.jsonrpc !== undefined && json.method === undefined) {
        await this._processJsonRpcResponse(json);
      }
    });

    this._process.stderr.on('data', (data) => {
      this._onErrorCallback(data);
    });

    this._process.on('close', (code) => {
      this._onCloseCallback(code);
    });
  }

  /**
   * @return {Process} process attached to the bridge
   */
  getProcess() {
    return this._process;
  }

  /**
   * sends a raw JSON message to attached process
   * @param {Object} msg - JSON to be sent
   */
  sendRaw(msg) {
    let base64data;
    try {
      msg = JSON.stringify(msg);
      const buff = Buffer.from(msg);
      base64data = buff.toString('base64');
    } catch (err) {
      throw new Error(`Cannot create ProcessRpcBridge message. Invalid data. ${err}`);
    }
    try {
      this._process[this._outputName].setEncoding('utf-8');
      this._process[this._outputName].write(base64data + '\n');
    } catch (err) {
      throw new Error(`Cannot send ProcessRpcBridge message. ${err}`);
    }
  }

  /**
   * Call method of RPC target whitin attached process
   * For example running `bridge.call('methodName', [1,2,3])` will call
   * `rpcTarget.methodName(1, 2, 3);`
   * @async
   * @param {String} methodName - name of method to be called
   * @param {Array} params - list of parameters to be passed to the method
   * @return {Promise}
   */
  call(methodName, params) {
    params = params || [];
    let id = '';
    while (id.length < 32) {
      id += Math.floor(Math.random()*0xffffffff).toString(16);
    }
    const msg = {
      jsonrpc: '2.0',
      method: methodName,
      params: params,
      id: id,
    };
    let promiseResolve;
    let promiseReject;
    const promise = new Promise((resolve, reject) => {
      promiseResolve = resolve;
      promiseReject = reject;
    });
    this._responseMap[id] = {
      promise: promise,
      resolve: promiseResolve,
      reject: promiseReject,
      stack: (new Error()).stack,
    };
    this.sendRaw(msg);

    const timeoutPromise = new Promise((resolve, reject) => {
      setTimeout(() => {
        reject(new Error('Timed out. No respone from the process in '+ this._timeout + 'ms.'));
      }, this._timeout);
    });

    return Promise.race([
      promise,
      timeoutPromise,
    ]);
  }

  /**
   * @param {Function} callback - function to be called when raw message is
   *        received from bridged process
   */
  onRaw(callback) {
    this._onRawMessageCallback = callback;
  }

  /**
   * @param {Function} callback - function to be called when attached process
   *        is closed
   */
  onClose(callback) {
    this._onCloseCallback = callback;
  }

  /**
   * @param {Function} callback - function to be called when attached process
   *        returns an error
   */
  onError(callback) {
    this._onErrorCallback = callback;
  }
}

module.exports = ProcessRpcBridge;
