/**
 * Handle errors from  service proces preserving stack trace
 */
class ProcessRpcBridgeError extends Error {
  /**
   * @constructor
   * @param {String} message - A human-readable description of the error.
   * @param {String} fileName - The value for the fileName property on the
   *        created Error objec
   * @param {String} lineNumber - The value for the lineNumber property on the
   *        created Error object
   */
  constructor(message, fileName, lineNumber) {
    super(message, fileName, lineNumber);
    this.stack = this._cleanupStack(this.stack);
  }

  /**
   * extends stack trace of the error by adding extra lines at the end
   * @param {String} extraStack - stacktrace (multi-line text)
   */
  extendStack(extraStack) {
    this.stack += '\n' + this._cleanupStack(extraStack);
  }

  /**
   * remove unnecessary lines from stacktrace (e.g. error message)
   * @param {String} stackTrace - stack trace to be clean up
   * @return {String} stack trace after clean up
   */
  _cleanupStack(stackTrace) {
    return stackTrace
        .split('\n')
        .filter((el) => {
          if (!el.match(/^\s*at/)) {
            return false;
          }
          if (el.match(/ProcessRpcBridge\.js/)) {
            return false;
          }
          if (el.match(/ProcessRpcBridgeError\.js/)) {
            return false;
          }
          return true;
        })
        .join('\n');
  }

  /**
   * Converts error to JSON-RPC response format
   * @param {String} id - ID used in JSON-RPC to pair request and response
   * @return {Object} JSON-RPC error object
   */
  toJsonRpc(id) {
    return {
      jsonrpc: '2.0',
      error: {
        code: -32603,
        message: this.message,
        data: {
          stack: this.stack,
          fileName: this.fileName,
          lineNumber: this.lineNumber,
        },
      },
      id: id,
    };
  }
}

ProcessRpcBridgeError.fromError = (err) => {
  const result = new ProcessRpcBridgeError(
      err.message,
      err.fileName,
      err.lineNumber
  );
  result.stack = '';
  result.extendStack(err.stack);
  result.extendStack((new Error()).stack);
  return result;
};

ProcessRpcBridgeError.fromJson = (json) => {
  json.error = json.error || {};
  json.error.code = json.error.code || -32603;
  json.error.message = json.error.message || 'unknown error';
  json.error.data = json.error.data || {};
  json.error.data.fileName = json.error.data.fileName || '';
  json.error.data.lineNumber = json.error.data.lineNumber || 0;
  json.error.data.stack = json.error.data.stack || '';
  const result = new ProcessRpcBridgeError(
      json.error.message,
      json.error.data.fileName,
      json.error.data.lineNumber
  );
  result.stack = '';
  result.extendStack(json.error.data.stack);
  result.extendStack((new Error()).stack);
  return result;
};

module.exports = ProcessRpcBridgeError;
