/**
 * RPC target for Process Bridge. It proxy all calls to HostedService exposing
 * requried methods
 * @memberof module:core
 */
class ServiceProxy {
  /**
   * @constructor
   * @param {HostedService} service - service where all calls should be passed
   * @param {ServiceRegistry} serviceRegistry - registry where the service is
   *        stored
   */
  constructor(service, serviceRegistry) {
    this._service = service;
    this._serviceRegistry = serviceRegistry;
  }

  /**
   * Calls service from the registry
   * @param {String} serviceName - query to select the service. For example:
   *        MyServiceName, MyServiceName?label=dev, ?label=dev
   * @param {String} methodName - name of method to be call
   * @param {Array} params - list of parameters to be passed to the method
   * @return {Object} response from the service
   */
  async callService(serviceName, methodName, params) {
    return await this._serviceRegistry.callService(
        serviceName,
        methodName,
        params
    );
  }

  /**
   * notify that service process is spawned
   */
  onProcessCreated() {
    this._service.onProcessCreated();
  }

  /**
   * @return {String} Name of the service. Generated automaticaly from its
   *         class name
   */
  getServiceName() {
    return this._service.name;
  }

  /**
   * @return {String} Unique ID of service instance
   */
  getServiceId() {
    return this._service.id;
  }

  /**
   * Log message with level fatal
   * @param {String} msg - message to be logged
   */
  logFatal(msg) {
    this._service.logger.fatal(msg);
  }

  /**
   * Log message with level error
   * @param {String} msg - message to be logged
   */
  logError(msg) {
    this._service.logger.error(String(msg));
  }

  /**
   * Log message with level warn
   * @param {String} msg - message to be logged
   */
  logWarn(msg) {
    this._service.logger.warn(String(msg));
  }

  /**
   * Log message with level info
   * @param {String} msg - message to be logged
   */
  logInfo(msg) {
    this._service.logger.info(String(msg));
  }

  /**
   * Log message with level debug
   * @param {String} msg - message to be logged
   */
  logDebug(msg) {
    this._service.logger.debug(String(msg));
  }
}

module.exports = ServiceProxy;
