const Service = require('../../Service.js');
const HttpServerRequest = require('./HttpServerRequest.js');
const http = require('http');

/**
 * Starts http server at given host and port.
 * All request can be forwarded to other micro services via routing rules
 *
 * #### Configuration
 *
 *   Option      | Default     | Description
 *  -------------|-------------|---------------
 *  `host`       | localhost   | Host where the server is listening. To bind to all network interfaces set to `0.0.0.0`
 *  `port`       | 8080        | Port where the server is listening
 *  `overflowLimit` | 1e6      | Maximum HTTP payload size
 *  `routing`   | []           | List of **Routing Rules**
 *
 * ##### Configuration: Routing Rules
 *
 *   Option      | Default     | Description
 *  -------------|-------------|---------------
 *  `prefix`     |             | Begining of URI used to match request to the rule. For example `/api`. If a request match several routing rules, the one with the longest prefix will be used. For example request to `/api/EchoService` may match both rules: `/` and `/api`. However `/api` is a longer prefix and thus it will be chosen.
 *  `service`    |             | Name of service to be called when routing rule is matched
 *  `method`     |             | Name of method to be called when routing rule is matched. it will get as an argument `HttpServerRequest` object and should return `HttpServerResponse`
 *  `dropPrefix` | true        | Whether full request URI should be included in `HttpServerRequest` or just its part after the prefix.
 *  `allowOrigin`|             | RegExp String for origins that allows CORS. For example: `'http:\\/\\/example.com'`,  `'http:\\/\\/127.0.0.1:9000'`, `'.*'`
 *
 * @memberof module:httpServer
 */
class HttpServerService extends Service {
  /**
   * initialize server and starts HTTP server
   */
  async init() {
    const port = this.config.port || 8080;
    const host = this.config.host || 'localhost';

    this._server = http.createServer((req, res) => this.request(req, res));
    this._server.on('error', (err) => {
      this.logError(String(err));
    });
    await new Promise((resolve) => {
      this._server.listen(port, host, resolve);
    });
    this.logInfo(`HTTP server started: http://${host}:${port}`);
    this.expose('request');
  }

  /**
   * process http request
   * @param {http.IncomingMessage} req - http request
   * @param {http.ServerResponse} res - http response
   */
  async request(req, res) {
    // // TODO: make sure that this is called from local container
    // // over the network callbacks of req/res will not work

    let routing = this.config.routing || {};
    const overflowLimit = this.config.overflowLimit || 1e6;

    return new Promise((resolve) => {
      this.logInfo(`HTTP ${req.method} ${req.url}`);
      if (Object.keys(routing).length == 0) {
        this.logWarn('No routing defined. Nothing will happen');
        res.writeHead(404);
        res.end('404 Not Found');
        return resolve();
      }

      routing = routing.sort((a, b) => {
        return b.prefix.length - a.prefix.length;
      });
      const reqData = new HttpServerRequest();
      reqData.url = req.url;
      reqData.method = req.method;
      reqData.headers = req.headers || {};
      reqData.httpVersion = req.httpVersion;
      reqData.body = '';
      let reqProcessError = false;
      req.on('data', (data) => {
        this.logDebug(`Processing chunk of request body (${data.length}B)`);
        reqData.body += data.toString('utf8');
        if (reqData.body.length > overflowLimit) {
          this.logError('Request payload exceedes limit of ' + overflowLimit + 'B');
          reqData._data = '';
          res.writeHead(413);
          res.end('413 Payload Too Large');
          req.connection.destroy();
          reqProcessError = true;
          return resolve();
        }
      });
      req.on('end', async () => {
        this.logDebug('Processing of request completed');
        if (reqProcessError) {
          this.logDebug('Stopping due to an error');
          return resolve();
        }

        const routingRule = this.resolveRouting(reqData.url);
        if (!routingRule) {
          this.logDebug('No routing found for' + reqData.url);
          res.writeHead(404);
          res.end('404 Not Found');
          return resolve();
        }
        if (!routingRule.service) {
          this.logError('no service defined for routing prefix' + routingRule.prefix);
          res.writeHead(500);
          res.end('500 Icorrect routing configuration');
          return resolve();
        }
        if (!routingRule.method) {
          this.logError('no method defined for routing prefix' + routingRule.prefix);
          res.writeHead(500);
          res.end('500 Icorrect routing configuration');
          return resolve();
        }
        if (routingRule.dropPrefix === undefined) {
          routingRule.dropPrefix = true;
        }
        if (routingRule.prefix === undefined) {
          routingRule.prefix = '/';
        }
        if (routingRule.dropPrefix) {
          reqData.url = reqData.url.substring(routingRule.prefix.length);
        }
        if (reqData.url[0] != '/') {
          reqData.url = '/' + reqData.url;
        }

        // CORS
        const isAllowed = this._isOriginAllowed(
            routingRule,
            reqData.headers.origin
        );

        // preflight
        if (
          reqData.method == 'OPTIONS' &&
          reqData.headers.origin !== undefined
        ) {
          this.logDebug(`Sending CORS preflight request. ${isAllowed ? 'Allow' : 'Deny'} origin:  ${reqData.headers.origin}`);
          if (isAllowed) {
            res.writeHead(200, {
              'Access-Control-Allow-Origin': reqData.headers.origin,
              'Access-Control-Allow-Headers': 'Content-Type',
            });
            res.end('200 OK');
            return resolve();
          } else {
            res.writeHead(403);
            res.end('403 Forbidden');
            return resolve();
          }
        }

        this.logDebug('Routing target lookup: ' + routingRule.service);

        this.logDebug('calling request handler...');
        let response;
        try {
          response = await this.callService(
              routingRule.service,
              routingRule.method,
              [reqData]
          );
        } catch (err) {
          this.logError(String(err));
          res.writeHead(500);
          res.end('500 Internal Error');
          return resolve();
        }

        response = response || {};
        this.logDebug('Request handling completed');
        response.headers = response.headers || {};
        if (reqData.headers.origin !== undefined && isAllowed) {
          response.headers['Access-Control-Allow-Origin'] = reqData.headers.origin;
        }
        res.writeHead(response.status || 200, response.headers);
        const body = response.binary ? Buffer.from(response.body, 'base64') : response.body;
        res.end(body || undefined);
        return resolve();
      });
    });
  }

  /**
   *
   * @param {Object} routingRule - routing rule with allowOrigin config
   * @param {String} origin - origin to be checked
   * @return {Boolean} whether the origin is allowed in routing rule or not
   * @private
   */
  _isOriginAllowed(routingRule, origin) {
    if (routingRule.allowOrigin === undefined) {
      this.logDebug(`CORS rule not defined in routing. Preflight rejected`);
      return false;
    }
    let pattern;
    try {
      pattern = new RegExp(routingRule.allowOrigin);
    } catch (err) {
      this.logError(`Cannot parse regexp pattern for allow origin. Rule: '${routingRule.prefix}'. RegExp: '${routingRule.allowOrigin}'`);
      this.logError(String(err));
      return false;
    }
    if (!pattern.test(origin)) {
      this.logDebug(`CORS rule not matched for '${routingRule.prefix}'. Origin '${origin}' does not match pattern '${routingRule.allowOrigin}'`);
      return false;
    }
    this.logDebug(`CORS rule matched! 200 OK`);
    return true;
  }

  /**
   * @param {String} url - url to be matched with routing rules
   * @return {Object} - routing configuration that match ptovided url
   */
  resolveRouting(url) {
    for (const rule of this.config.routing) {
      if (!rule.prefix || url.substring(0, rule.prefix.length) == rule.prefix) {
        return rule;
      }
    }
    return null;
  }
}

module.exports = HttpServerService;
