const Service = require('../../Service.js');
const HttpServerResponse = require('../httpServer/HttpServerResponse.js');
const fs = require('fs');

/**
 * Provides client of JSON-RPC (Frontend JS lib, or web client)
 *
 * #### Configuration
 *
 *   Option          | Default     | Description
 *  -----------------|-------------|---------------
 *  `defaultGateway` | ""          | URI of gateway if none is provided (e.g. /api).
 *
 * @memberof module:jsonRpcClient
 */
class JsonRpcClientService extends Service {
  /**
   * initialize the service exposing services
   */
  async init() {
    this._clientLibSource = __dirname + '/client_lib.babel.js_';
    this._webClientSource = __dirname + '/web_client.html';
    this.config.defaultGateway = this.config.defaultGateway || '';
    this.expose('getWebClient');
    this.expose('getClientLib');
  }

  /**
   * @async
   * @return {HttpServerResponse} HTTP response containing web client
   */
  getWebClient() {
    return new Promise(async (resolve) => {
      const lib = await this.getClientLib();
      fs.readFile(this._webClientSource, 'utf8', (err, content) => {
        const response = new HttpServerResponse();
        if (err) {
          this.logError(err);
          response.status = 500;
          response.body = '500 Internal Error';
          return resolve(response);
        }
        response.headers = {
          'Content-Type': 'text/html; charset=utf-8',
        };
        content = content.replace(
            /{{GATEWAY_URL}}/g,
            this.config.defaultGateway
        );
        content = content.replace(
            /{{CLIENT_LIB}}/g,
            lib.body
        );
        response.body = content;
        return resolve(response);
      });
    });
  }

  /**
   * @async
   * @return {HttpServerResponse} HTTP response containing JS client lib
   */
  getClientLib() {
    return new Promise((resolve) => {
      fs.readFile(this._clientLibSource, 'utf8', (err, content) => {
        const response = new HttpServerResponse();
        if (err) {
          this.logError(err);
          response.status = 500;
          response.body = '500 Internal Error';
          return resolve(response);
        }
        response.headers = {
          'Content-Type': 'text/javascript;charset=UTF-8',
        };
        content = content.replace(
            /{{GATEWAY_URL}}/g,
            this.config.defaultGateway
        );
        response.body = content;
        return resolve(response);
      });
    });
  }
}

module.exports = JsonRpcClientService;
