/**
 * Client of JSON-RPC
 * @memberof module:jsonRpcClient
 */
class MicrosetJsonRpcClient {
  /**
   * @constructor
   */
  constructor() {
    this.defaultGateway = '{{GATEWAY_URL}}';
  }

  /**
   * Call defined service. It supports both: Promises and callbacks. If no
   * callback is provided, then promise is returned
   * @async
   * @param {String} service - name of service to be called
   * @param {String} method - name of method to be called
   * @param {Array} params - parameters to be passed to the service
   * @param {String} [gateway] - (Optional) gateway URL where the service is
   *        located
   * @param {String} [callback] - (Optional) callback to be called when the
   *        response is received: `function(err, response)`
   * @return {Object} Promise with response if no callback given
   */
  callService(service, method, params, gateway, callback) {
    if (!callback || typeof(gateway) == 'cd') {
      callback = gateway;
      gateway = undefined;
    }
    gateway = gateway || this.defaultGateway;
    if (gateway[gateway.length - 1] != '/') {
      gateway += '/';
    }
    if (callback) {
      this._callServicewithCallback(service, method, params, gateway, callback);
    } else {
      return new Promise((resolve, reject) => {
        this._callServicewithCallback(
            service,
            method,
            params,
            gateway,
            (err, result) => {
              if (err) {
                return reject(err);
              }
              return resolve(result);
            }
        );
      });
    }
  }

  /**
   * Call defined service. It supports both callbacks only
   * @async
   * @param {String} service - name of service to be called
   * @param {String} method - name of method to be called
   * @param {Array} params - parameters to be passed to the service
   * @param {String} gateway - gateway URL where the service is
   *        located
   * @param {String} callback - callback to be called when the
   *        response is received: `function(err, response)`
   * @private
   */
  _callServicewithCallback(service, method, params, gateway, callback) {
    const url = gateway + service;
    const xhr = new XMLHttpRequest();
    xhr.onload = function() {
      let outputJson;
      try {
        outputJson = JSON.parse(xhr.response);
      } catch (err) {
        return callback(new Error('Cannot parse JSON response: ' + xhr.response));
      }
      if (xhr.status >= 200 && xhr.status < 300) {
        return callback(null, outputJson.result);
      } else {
        return callback(new Error(outputJson.error.message));
      }
    };
    xhr.ontimeout = function() {
      return callback(new Error('Connection timeout on ' + url));
    };
    xhr.onerror = function() {
      return callback(new Error('Connection error on ' + url));
    };
    xhr.open('POST', url);
    xhr.setRequestHeader('Content-Type', 'application/json');
    const body = {
      jsonrpc: '2.0',
      method: method,
      params: params,
    };
    xhr.send(JSON.stringify(body));
  }
}

if (!window.microset) {
  window.microset = {};
}

window.microset.jsonRpcClient = new MicrosetJsonRpcClient();
