const HttpServerResponse = require('../httpServer/HttpServerResponse.js');

/**
 * JSON-RPC errors
 * @memberof module:jsonRpcGateway
 */
class JsonRpcError {
  /**
   * @constructor
   * @param {Number} httpCode - HTTP status code (e.g. 200)
   * @param {Number} jrpcCode - error codes of JSON-RPC protocol
   * @param {String} msg - error message to be displayed
   */
  constructor(httpCode, jrpcCode, msg) {
    this._httpCode = httpCode;
    this._jrpcCode = jrpcCode;
    this._msg = msg;
  }

  /**
   * @return {HttpResponse} - Http reponse represeting the error
   */
  toHttpResponse() {
    const result = new HttpServerResponse();
    result.headers = {
      'Content-Type': 'application/json',
      'Accept': 'application/json',
    };
    const body = {
      jsonrpc: '2.0',
      error: {
        'code': this._jrpcCode,
        'message': this._httpCode + ' ' + this._msg,
      },
    };
    result.body = JSON.stringify(body);
    result.status = this._httpCode;
    return result;
  }
}


JsonRpcError.methodNotAllowed = function() {
  return new JsonRpcError(405, -32000, 'Method Not Allowed. Only POST is supported.');
};

JsonRpcError.invalidJsonPayload = function() {
  return new JsonRpcError(400, -32700, 'Invalid JSON Payload');
};

JsonRpcError.invalidJsonRpc = function(msg) {
  return new JsonRpcError(400, -32600, 'JSON payload is not a valid request object. ' + msg);
};

JsonRpcError.serviceNotFound = function(serviceName) {
  return new JsonRpcError(404, -32601, 'Service \'' + serviceName + '\' not found');
};

JsonRpcError.serviceNotReady = function(serviceName) {
  return new JsonRpcError(504, -32601, 'Service \'' + serviceName + '\' is not ready. Try again later.');
};

JsonRpcError.methodNotFound = function(serviceName, methodName) {
  return new JsonRpcError(404, -32601, 'Method \'' + methodName + '\' not found in \'' + serviceName + '\'');
};

JsonRpcError.invalidParameters = function(msg) {
  return new JsonRpcError(400, -32602, 'Invalid method parameters. ' + msg);
};

JsonRpcError.forbidden = function(serviceName, methodName) {
  return new JsonRpcError(403, -32600, 'Method \'' + serviceName + '.' + methodName + '(...)\' forbidden');
};

JsonRpcError.internalError = function(msg) {
  msg = msg || '';
  return new JsonRpcError(500, -32603, msg || 'Internal JSON-RPC error.');
};


module.exports = JsonRpcError;
