const Service = require('../../Service.js');
const JsonRpcError = require('./JsonRpcError.js');
const HttpServerResponse = require('../httpServer/HttpServerResponse.js');

/**
 * Map JSON-RPC calls to micro services calls.
 *
 * #### Configuration
 *
 *   Option          | Default    | Description
 *  -----------------|------------|---------------
 *  `whitelist`      |            | List of **service** objects that are accessible over the gateway. If not provided, all services are available
 *
 * ##### Configuration: Service
 *
 *   Option      | Default     | Description
 *  -------------|-------------|---------------
 *  `service`    | *           | name of service that should be available
 *  `method`     | *           | name of method that should be available
 *
 * @memberof module:jsonRpcGateway
 */
class JsonRpcGatewayService extends Service {
  /**
   * Initialize the service
   */
  async init() {
    if (!this.config.whitelist || !Array.isArray(this.config.whitelist )) {
      this.config.whitelist = [
        {service: '*', method: '*'},
      ];
    }
    this.expose('request');
  }

  /**
   * Check if the service is whitelisted
   * @private
   * @param {String} serviceName - name of the service to be checked
   * @param {String} methodName - name of the method to be checked
   * @return {Boolean} true if the method is allowed, otherwise false
   */
  _isWhitelisted(serviceName, methodName) {
    for (const rule of this.config.whitelist) {
      const serviceMatch = (rule.service == serviceName || rule.service === '*');
      const methodMatch = (rule.method == methodName || rule.method === '*');
      if (serviceMatch && methodMatch) {
        return true;
      }
    }
    return false;
  }

  /**
   * Handle HTTP request by calling related service
   * @param {HttpRequest} request - HTTP Request
   * @return {HttpResponse} - response from the service or an error
   */
  async request(request) {
    // check if proper method was used
    if (request.method != 'POST') {
      return JsonRpcError.methodNotAllowed().toHttpResponse();
    }

    // get the endpoint
    const endpoint = request.url;

    // read the payload
    let queryData;
    try {
      queryData = JSON.parse(request.body);
    } catch (err) {
      return JsonRpcError.invalidJsonPayload().toHttpResponse();
    }

    if (!queryData.jsonrpc) {
      return JsonRpcError.invalidJsonRpc(
          '\'jsonrpc\' field is missing!'
      ).toHttpResponse();
    }

    if (queryData.jsonrpc != '2.0') {
      return JsonRpcError.invalidJsonRpc(
          `JSON-RPC version '${queryData.jsonrpc}' is not supported`
      ).toHttpResponse();
    }

    if (!queryData.method) {
      return JsonRpcError.invalidJsonRpc(
          '\'method\' field is missing!'
      ).toHttpResponse();
    }

    if (queryData.params && typeof(queryData.params) != 'object') {
      return JsonRpcError.invalidJsonRpc(
          `'params' must be an Array or Object`
      ).toHttpResponse();
    }

    const id = queryData.id;
    const serviceName = endpoint.replace(/^\//, '');
    const method = queryData.method;
    const params = queryData.params ? queryData.params : [];

    if (!this._isWhitelisted(serviceName, method)) {
      this.logDebug(`Method '${serviceName}.${method}(...)' forbidden`);
      return JsonRpcError.forbidden(serviceName, method).toHttpResponse();
    }

    try {
      const result = await this.callService(
          serviceName,
          method,
          Object.values(params)
      );
      const response = new HttpServerResponse();
      response.headers = {
        'Content-Type': 'application/json',
        'Accept': 'application/json',
      };
      const body = {
        jsonrpc: '2.0',
        result: result,
      };
      if (id !== undefined) {
        body.id = id;
      }
      response.body = JSON.stringify(body);
      response.status = 200;
      return response;
    } catch (err) {
      this.logError(`Internal error of ${serviceName}.${method}: ${err}`);
      return JsonRpcError.internalError(err.toString()).toHttpResponse();
    }
  }
}

module.exports = JsonRpcGatewayService;
