
const fs = require('fs');
const path = require('path');
const Service = require('../../Service.js');

/**
 * Simple file storage for JSON objects. It does not support clustering.
 *
 * #### Configuration
 *
 *   Option          | Default            | Description
 *  -----------------|--------------------|---------------
 *  `dataPath`       | json-storage-data  | Directory where to store json data
 *
 * @memberof module:jsonStorage
 */
class JsonStorageService extends Service {
  /**
   * Inisitalize the service
   */
  async init() {
    this.config.dataPath = this.config.dataPath || path.resolve(__dirname, '../../../json-storage-data/');
    this.expose('createItem');
    this.expose('updateItem');
    this.expose('getItem');
    this.expose('removeItem');
  }

  /**
   * Remove file at provided path
   * @param {String} filePath - path to file to be deleted
   * @return {Null} null
   * @async
   * @private
   */
  _removeFile(filePath) {
    return new Promise((resolve, reject) => {
      fs.unlink(filePath, (err) => {
        if (err) {
          return reject(err);
        }
        resolve();
      });
    });
  }

  /**
   * write data to file at provided path
   * @param {String} filePath - path to file to write
   * @param {String} data - data to be saved
   * @return {Null} null
   * @async
   * @private
   */
  _writeFile(filePath, data) {
    return new Promise((resolve, reject) => {
      fs.writeFile(filePath, data, (err) => {
        if (err) {
          return reject(err);
        }
        resolve();
      });
    });
  }

  /**
   * Create directory recuresly
   * @param {String} dirPath - path to dir to be created
   * @return {Null} null
   * @async
   * @private
   */
  _createDir(dirPath) {
    return new Promise((resolve, reject) => {
      fs.mkdir(dirPath, {recursive: true}, (err) => {
        if (err) {
          return reject(err);
        }
        resolve();
      });
    });
  }

  /**
   * read file at provided path
   * @param {String} filePath - path to file to be read
   * @return {String} content of the file
   * @async
   * @private
   */
  _readFile(filePath) {
    return new Promise((resolve, reject) => {
      fs.readFile(filePath, 'utf-8', (err, content) => {
        if (err) {
          return reject(err);
        }
        resolve(content);
      });
    });
  }

  /**
   * check if a file exists
   * @param {String} filePath - path to file to checked
   * @return {Boolean} true if the file exists, otherwise false
   * @async
   * @private
   */
  _exists(filePath) {
    return new Promise((resolve, reject) => {
      fs.exists(filePath, (result) => {
        resolve(result);
      });
    });
  }

  /**
   * Save a new item to the storage
   * @param {String} bucket - bucket where the data should be stored
   * @param {String} data - data to be stored
   * @return {String} id of created item
   */
  async createItem(bucket, data) {
    bucket = bucket ? String(bucket) : '';
    if (!bucket) throw new Error('Bucket name is required!');
    if (!data) throw new Error('Data object is required!');
    if (!(/^[A-Za-z0-9_]+$/).test(bucket)) throw new Error(`Invalid bucket name '${bucket}'`);
    if (typeof(data) != 'object') throw new Error('Invalid data argument. Data must be an object');

    let id = '';
    for (let i=0; i < 4; i++) {
      id += Math.floor(Math.random()*0xffff).toString(16);
    }
    while (id.length < 16) {
      id = '0' + id;
    }
    const filepath = path.join(this.config.dataPath, bucket, id + '.json');
    const dirpath = path.dirname(filepath);
    try {
      if (await this._exists(dirpath) == false) {
        this.logDebug(`Directory ${dirpath} does not exists`);
        await this._createDir(dirpath);
        this.logDebug(`Directory ${dirpath} created`);
      }
      await this._writeFile(filepath, JSON.stringify(data));
    } catch (err) {
      this.logError(String(err));
      throw new Error(`Cannot create item '${bucket}:${id}'`);
    }
    return id;
  }

  /**
   * updates existing item in the storage
   * @param {String} bucket - bucket where the data is located
   * @param {String} id of stored item
   * @param {String} data - new data to be stored
   */
  async updateItem(bucket, id, data) {
    bucket = bucket ? String(bucket) : '';
    id = id ? String(id) : '';
    if (!bucket) throw new Error('Bucket name is required!');
    if (!data) throw new Error('Data object is required!');
    if (!id) throw new Error('Item ID is required!');
    if (!(/^[A-Za-z0-9_]+$/).test(bucket)) throw new Error(`Invalid bucket name '${bucket}'`);
    if (!(/^[0-9a-f]+$/).test(id)) throw new Error(`Invalid item id '${bucket}'`);
    if (typeof(data) != 'object') throw new Error('Invalid data argument. Data must be an object');

    const filepath = path.join(this.config.dataPath, bucket, id + '.json');
    const exists = await this._exists(filepath);
    if (!exists) {
      this.logDebug(`Item '${bucket}:${id}' not found`);
      throw new Error(`Item '${bucket}:${id}' not found`);
    }
    try {
      await this._writeFile(filepath, JSON.stringify(data));
    } catch (err) {
      this.logError(String(err));
      throw new Error(`Cannot update item '${bucket}:${id}'`);
    }
  }

  /**
   * gets an item from the storage
   * @param {String} bucket - bucket where the data is located
   * @param {String} id of stored item
   * @return {Object} stored data
   */
  async getItem(bucket, id) {
    bucket = bucket ? String(bucket) : '';
    id = id ? String(id) : '';
    if (!bucket) throw new Error('Bucket name is required!');
    if (!id) throw new Error('Item ID is required!');
    if (!(/^[A-Za-z0-9_]+$/).test(bucket)) throw new Error(`Invalid bucket name '${bucket}'`);
    if (!(/^[0-9a-f]+$/).test(id)) throw new Error(`Invalid item id '${bucket}'`);

    const filepath = path.join(this.config.dataPath, bucket, id + '.json');
    const exists = await this._exists(filepath);
    if (!exists) {
      this.logDebug(`Item '${id}' not found`);
      throw new Error(`Item '${id}' not found`);
    }
    let content;
    try {
      content = await this._readFile(filepath);
    } catch (err) {
      this.logDebug(String(err));
      throw new Error(`Cannot read item '${bucket}:${id}'`);
    }
    try {
      content = JSON.parse(content);
    } catch (err) {
      this.logDebug(String(err));
      throw new Error(`Unable to parse content of item '${bucket}:${id}'`);
    }
    return content;
  }

  /**
   * remove item in from the storage
   * @param {String} bucket - bucket where the data is located
   * @param {String} id of stored item
   */
  async removeItem(bucket, id) {
    bucket = bucket ? String(bucket) : '';
    id = id ? String(id) : '';
    if (!bucket) throw new Error('Bucket name is required!');
    if (!id) throw new Error('Item ID is required!');
    if (!(/^[A-Za-z0-9_]+$/).test(bucket)) throw new Error(`Invalid bucket name '${bucket}'`);
    if (!(/^[0-9a-f]+$/).test(id)) throw new Error(`Invalid item id '${bucket}'`);

    const filepath = path.join(this.config.dataPath, bucket, id + '.json');
    const exists = await this._exists(filepath);
    if (!exists) {
      this.logDebug(`Item '${id}' not found`);
      throw new Error(`Item '${id}' not found`);
    }
    try {
      await this._removeFile(filepath);
    } catch (err) {
      this.logDebug(String(err));
      throw new Error(`Cannot remove item '${bucket}:${id}'`);
    }
  }
}

module.exports = JsonStorageService;
