const Service = require('../../Service.js');
const path = require('path');
const fs = require('fs');
const HttpServerResponse = require('../httpServer/HttpServerResponse.js');

/**
 * Provides static files for HttpServerService
 *
 * #### Configuration
 *
 *   Option          | Default    | Description
 *  -----------------|------------|---------------
 *  `webRoot`        | public     | path to static files to be hosted
 *
 * @memberof module:staticContent
 */
class StaticContentService extends Service {
  /**
   * Initialize the service
   */
  async init() {
    this.config.webRoot = this.config.webRoot || path.resolve(__dirname, '../../../public');
    this.config.webRoot = path.resolve(this.config.webRoot);
    this.expose('readFile');
  }

  /**
   * read file at provided path
   * @param {String} filePath - path to file to be read
   * @return {String} content of the file
   * @async
   * @private
   */
  _readFile(filePath) {
    return new Promise((resolve, reject) => {
      fs.readFile(filePath, (err, content) => {
        if (err) {
          return reject(err);
        }
        resolve(content.toString('base64'));
      });
    });
  }

  /**
   * check if a file exists (resolving index.html if needed)
   * @param {String} filePath - path to file to checked
   * @return {Boolean} true if the file exists, otherwise false
   * @async
   * @private
   */
  _resolve(filePath) {
    return new Promise((resolve, reject) => {
      fs.exists(filePath, (exists) => {
        if (!exists) {
          this.logDebug(`File '${filePath}' not found`);
          return resolve(null);
        }
        fs.stat(filePath, (err, stats) => {
          if (err) {
            this.logError(err);
            return resolve(null);
          }
          if (stats.isDirectory()) {
            filePath = path.join(filePath, 'index.html');
          }
          fs.exists(filePath, (exists) => {
            if (!exists) {
              this.logDebug(`File '${filePath}' not found`);
              return resolve(null);
            }
            return resolve(filePath);
          });
        });
      });
    });
  }

  /**
   * @param {String} ext - file extension (e.g. `.html`)
   * @return {String} coresponding mimetype
   * @private
   */
  _getMimeType(ext) {
    const mimes = [];
    mimes['.aac'] = 'audio/aac';
    mimes['.abw'] = 'application/x-abiword';
    mimes['.arc'] = 'application/x-freearc';
    mimes['.avi'] = 'video/x-msvideo';
    mimes['.azw'] = 'application/vnd.amazon.ebook';
    mimes['.bin'] = 'application/octet-stream';
    mimes['.bmp'] = 'image/bmp';
    mimes['.bz'] = 'application/x-bzip';
    mimes['.bz2'] = 'application/x-bzip2';
    mimes['.csh'] = 'application/x-csh';
    mimes['.css'] = 'text/css';
    mimes['.csv'] = 'text/csv';
    mimes['.doc'] = 'application/msword';
    mimes['.docx'] = 'application/vnd.openxmlformats-officedocument.wordprocessingml.document';
    mimes['.eot'] = 'application/vnd.ms-fontobject';
    mimes['.epub'] = 'application/epub+zip';
    mimes['.gif'] = 'image/gif';
    mimes['.htm'] = 'text/html';
    mimes['.html'] = 'text/html';
    mimes['.ico'] = 'image/vnd.microsoft.icon';
    mimes['.ics'] = 'text/calendar';
    mimes['.jar'] = 'application/java-archive';
    mimes['.jpeg'] = 'image/jpeg';
    mimes['.jpg'] = 'image/jpeg';
    mimes['.js'] = 'text/javascript';
    mimes['.json'] = 'application/json';
    mimes['.jsonld'] = 'application/ld+json';
    mimes['.mid'] = 'audio/midi';
    mimes['.midi'] = 'audio/midi';
    mimes['.mjs'] = 'text/javascript';
    mimes['.mp3'] = 'audio/mpeg';
    mimes['.mpeg'] = 'video/mpeg';
    mimes['.mpkg'] = 'application/vnd.apple.installer+xml';
    mimes['.odp'] = 'application/vnd.oasis.opendocument.presentation';
    mimes['.ods'] = 'application/vnd.oasis.opendocument.spreadsheet';
    mimes['.odt'] = 'application/vnd.oasis.opendocument.text';
    mimes['.oga'] = 'audio/ogg';
    mimes['.ogv'] = 'video/ogg';
    mimes['.ogx'] = 'application/ogg';
    mimes['.otf'] = 'font/otf';
    mimes['.png'] = 'image/png';
    mimes['.pdf'] = 'application/pdf';
    mimes['.ppt'] = 'application/vnd.ms-powerpoint';
    mimes['.pptx'] = 'application/vnd.openxmlformats-officedocument.presentationml.presentation';
    mimes['.rar'] = 'application/x-rar-compressed';
    mimes['.rtf'] = 'application/rtf';
    mimes['.sh'] = 'application/x-sh';
    mimes['.svg'] = 'image/svg+xml';
    mimes['.swf'] = 'application/x-shockwave-flash';
    mimes['.tar'] = 'application/x-tar';
    mimes['.tif'] = 'image/tiff';
    mimes['.tiff'] = 'image/tiff';
    mimes['.ttf'] = 'font/ttf';
    mimes['.txt'] = 'text/plain';
    mimes['.vsd'] = 'application/vnd.visio';
    mimes['.wav'] = 'audio/wav';
    mimes['.weba'] = 'audio/webm';
    mimes['.webm'] = 'video/webm';
    mimes['.webp'] = 'image/webp';
    mimes['.woff'] = 'font/woff';
    mimes['.woff2'] = 'font/woff2';
    mimes['.xhtml'] = 'application/xhtml+xml';
    mimes['.xls'] = 'application/vnd.ms-excel';
    mimes['.xlsx'] = 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet';
    mimes['.xml'] = 'application/xml';
    mimes['.xul'] = 'application/vnd.mozilla.xul+xml';
    mimes['.zip'] = 'application/zip';
    mimes['.3gp'] = 'video/3gpp';
    mimes['.3g2'] = 'video/3gpp2';
    mimes['.7z'] = 'application/x-7z-compressed';
    return mimes[ext] ? mimes[ext] : null;
  }

  /**
   * @param {HttpRequest} request - HTTP Request to static file
   * @return {HttpResponse} - response from the service or an error
   */
  async readFile(request) {
    const response = new HttpServerResponse();
    let url = request.url;
    url = url.replace(/\?.*$/, '');
    url = url.replace(/\#.*$/, '');
    url = url.replace(/^\/*/, '');
    let filepath = path.resolve(this.config.webRoot, url);
    const pathPrefix = filepath.substring(0, this.config.webRoot.length);
    if (pathPrefix != this.config.webRoot) {
      this.logDebug(`Request '${request.url}' is forbidden. File '${filepath}' is outside of web root '${this.config.webRoot}'`);
      response.status = 403;
      response.body = '403 Forbidden';
      return response;
    }
    filepath = await this._resolve(filepath);
    if (!filepath) {
      this.logDebug(`File '${request.url}' not found`);
      response.status = 404;
      response.body = '404 Not Found';
      return response;
    }
    let content;
    try {
      content = await this._readFile(filepath);
    } catch (err) {
      this.logDebug(String(err));
      this.logDebug(`Cannot read file '${filepath}'`);
      response.status = 500;
      response.body = '500 Internal Error';
      return response;
    }
    response.status = 200;
    response.body = content;
    const contentType = this._getMimeType(path.extname(filepath));
    if (contentType) {
      response.headers = {
        'Content-Type': contentType,
      };
    }
    response.binary = true;
    return response;
  }
}

module.exports = StaticContentService;
